/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package lp;

import ilog.concert.IloException;
import ilog.concert.IloNumExpr;
import ilog.concert.IloNumVar;
import ilog.cplex.IloCplex;
import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import static java.awt.RenderingHints.KEY_ANTIALIASING;
import static java.awt.RenderingHints.VALUE_ANTIALIAS_ON;
import javax.swing.JFrame;
import javax.swing.JPanel;

/**
 *
 * @author marci
 */
public class LP {
    private static final int D = 2;         //considering a 2D scenario
    private static final int N = 50;         //5 time steps
    private static final double dt = 0.1;   //1 second for each time step
    private static final double Umax = 1.0; //maximum control to be applied
    
    //Dynamic matrixes
    private static final double A[][] = new double[][]{
        {1      ,0      ,dt     ,0},
        {0      ,1      ,0      ,dt},
        {0      ,0      ,1      ,0},
        {0      ,0      ,0      ,1},
    };
    private static final double B[][] = new double[][]{
        {dt*dt/2    ,0},
        {0          ,dt*dt/2},
        {dt         ,0},
        {0          ,dt},
    };
    
    //spatial restrictions
    private static final double H[][] = new double[][]{
        {1      ,0},
        {0      ,1},
        {-1     ,0},
        {0      ,-1},
    };
    private static final double g[] = new double[]{
        1,
        1,
        1,
        1
    };
    
    //initial conditions
    //private static final double start[] = new double[]{-0.5, -0.5, 0, 0};         //stopped
    private static final double start[] = new double[]{-0.5, -0.5, 0, 1.5};           //to up
    

    //end conditions
    //private static final double end[] = new double[]{0.5, 0.5};                     //inertial
    private static final double end[] = new double[]{0.5, 0.5, 1, 0};             //to right
    
    public static void main(String[] args) throws IloException {
        // Create the enviroment
        IloCplex cplex = new IloCplex();
        
        // Define the control variables
        IloNumVar u[][] = new IloNumVar[N-1][D];
        for(int n=0; n<N-1; n++){
            for(int d=0; d<D; d++){
                u[n][d] = cplex.numVar(-Umax, +Umax, "u("+n+","+d+")");
            }
        }
        
        // Define the state variables
        IloNumVar x[][] = new IloNumVar[N][2*D];
        for(int n=0; n<N; n++){
            for(int d=0; d<2*D; d++){
                x[n][d] = cplex.numVar(Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, "x("+n+","+d+")");
            }
        }
        
        // Objective function: minimize g(.) = sum_n { u(n)*u(n) }
        IloNumExpr obj = null;
        for(int n=0; n<N-1; n++){
            for(int d=0; d<D; d++){
                if(obj==null){
                    obj = cplex.prod(u[n][d], u[n][d]);
                }else{
                    obj = cplex.sum(obj, cplex.prod(u[n][d], u[n][d]));
                }
            }
        }
        cplex.addMinimize(obj);
        
        // Dynamic equations: x(n+1) = A x(n) + B u(n)
        for(int n=0; n<N-1; n++){
            for(int d=0; d<2*D; d++){
                IloNumExpr expr = null;
                //A x(n)
                for(int j=0; j<2*D; j++){
                    if(expr==null){
                        expr = cplex.prod(A[d][j], x[n][j]);
                    }else{
                        expr = cplex.sum(expr, cplex.prod(A[d][j], x[n][j]));
                    }
                }
                //B u(n)
                for(int j=0; j<D; j++){
                    expr = cplex.sum(expr, cplex.prod(B[d][j], u[n][j]));
                }
                cplex.addEq(x[n+1][d], expr, "dy("+n+","+d+")");
            }
        }
        //spatial restrictions: H x(n) <= g
        for(int n=0; n<N; n++){
            for(int i=0; i<H.length; i++){
                IloNumExpr expr = null;
                //h(i) x(n)
                for(int d=0; d<D; d++){
                    if(expr==null){
                        expr = cplex.prod(H[i][d], x[n][d]);
                    }else{
                        expr = cplex.sum(expr, cplex.prod(H[i][d], x[n][d]));
                    }
                }
                cplex.addLe(expr, g[i], "sr("+n+","+i+")");
            }
        }
        
        // Initial conditions: x(0) = start
        for(int d=0; d<start.length; d++){
            x[0][d].setLB(start[d]);
            x[0][d].setUB(start[d]);
        }
        // End conditions: x(N-1) = end
        for(int d=0; d<end.length; d++){
            x[N-1][d].setLB(end[d]);
            x[N-1][d].setUB(end[d]);
        }
        
        cplex.exportModel("./model.lp");
        
        if(cplex.solve()){
            System.out.println("Solved, status  = "+cplex.getStatus());
            System.out.println("Objective       = "+cplex.getObjValue());
            
            //getting solution
            final double X[][] = new double[N][];
            for(int n=0; n<N; n++){
                X[n] = cplex.getValues(x[n]);
            }
            final double U[][] = new double[N-1][];
            for(int n=0; n<N-1; n++){
                U[n] = cplex.getValues(u[n]);
            }
            System.out.printf("%2s | %9s %9s %9s %9s | %9s %9s\n", "n", "px", "py", "vx", "vy", "ux", "uy");
            for(int n=0; n<N; n++){
                System.out.printf("%2d | %9.2f %9.2f %9.2f %9.2f | %9.2f %9.2f\n", 
                        n, X[n][0], X[n][1], X[n][2], X[n][3], 
                        n<N-1? U[n][0] : Double.NaN, n<N-1 ? U[n][1] : Double.NaN);
            }
            
            
            // --------------------------- plot ---------------------------
            final int M = 250;
            final int W = 800;
            final int H = 600;
            
            JFrame frame = new JFrame("LP");
            frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
            frame.setSize(W+16, H+40);
            JPanel draw = new JPanel(){
                @Override
                protected void paintComponent(Graphics g) {
                    super.paintComponent(g); //To change body of generated methods, choose Tools | Templates.
                    
                    Graphics2D g2 = (Graphics2D) g;
                    g2.setRenderingHint(KEY_ANTIALIASING, VALUE_ANTIALIAS_ON);
                    g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 1.0f));
             
                    g.setColor(Color.WHITE); 
                    g.fillRect(0, 0, W, H);
                    
                    g.translate(W/2, H/2);
                    
                    g.setColor(Color.BLACK); 
                    g.drawRect(-M, -M, 2*M, 2*M);
                    
                    g.fillOval(c(start[0])-5, c(-start[1])-5, 10, 10);
                    
                    g.fillRect(c(end[0])-5, c(-end[1])-5, 10, 10);
                    
                    for(int n=0; n<X.length; n++){
                        g.fillOval(c(X[n][0])-3, c(-X[n][1])-3, 6, 6);
                    }
                    
                    
                    //rota com ruido
                    double Xreal[][] = new double[N][2*D];
                    for(int n=0; n<Xreal.length; n++){
                        for(int j=0; j<2*D; j++){
                            if(n==0){
                                Xreal[n][j] = start[j];
                            }else{
                                Xreal[n][j] = 0;
                                for(int i=0; i<2*D; i++){
                                    Xreal[n][j] += A[j][i] * Xreal[n-1][i];
                                }
                                for(int i=0; i<D; i++){
                                    Xreal[n][j] += B[j][i] * U[n-1][i];
                                }
                                for(int i=0; i<D; i++){
                                    Xreal[n][j] += -0.005 + Math.random()*0.01;
                                }
                            }
                        }
                    }
                    
                    g.setColor(Color.red);
                    for(int n=0; n<Xreal.length; n++){
                        g.fillOval(c(Xreal[n][0])-3, c(-Xreal[n][1])-3, 6, 6);
                    }
                }
                private int c(double x){
                    return (int)(x*M);
                }
            };
            frame.add(draw);
            frame.setVisible(true);
            
        }else{
            System.out.println("Not solved, status = "+cplex.getStatus());
        }
    }
    
}
